# encoding: utf-8
# Copyright (C) 2015 John Törnblom
'''
Keep version info in one single place
'''


name = 'pyrsl'
date = '2015-05-05'
release = '0.3.1'

complete_string = '%s v%s (%s)' % (name, release, date)
